@@ -1,19 +1,24 @@
 # -*- coding: utf-8 -*-
+from __future__ import print_function
+
 
 # C Bindings
 
-import os, sys, functools as ft
+import os, sys, ctypes.util, functools as ft
 from ctypes import *
 
 
-def force_str(s): return s.decode('utf-8') if isinstance(s, bytes) else s
-def force_bytes(s): return s.encode('utf-8') if isinstance(s, unicode) else s
+force_str = lambda s, errors='strict': s.decode('utf-8', errors) if isinstance(s, bytes) else s
+force_bytes = lambda s, errors='strict': s.encode('utf-8', errors) if isinstance(s, unicode) else s
 
 if sys.version_info.major >= 3:
 	class c_str_p_type(object):
 		c_type = c_char_p
 		def __call__(self, val): return force_str(val)
-		def from_param(self, val): return force_bytes(val)
+		def from_param(self, val):
+			# int will be interpreted as pointer and segfault in py3
+			if isinstance(val, int): raise ArgumentError(type(val))
+			return force_bytes(val)
 	unicode, c_str_p = str, c_str_p_type()
 
 	import time
@@ -37,6 +42,8 @@ else:
 		return ts.tv_sec + ts.tv_nsec * 1e-9
 
 
+PA_INVALID = 2**32-1
+
 PA_VOLUME_NORM = 0x10000
 PA_VOLUME_MAX = 2**32-1 // 2
 PA_VOLUME_INVALID = 2**32-1
@@ -88,10 +95,43 @@ PA_SUBSCRIPTION_EVENT_CHANGE = 0x0010
 PA_SUBSCRIPTION_EVENT_REMOVE = 0x0020
 PA_SUBSCRIPTION_EVENT_TYPE_MASK = 0x0030
 
+PA_SAMPLE_FLOAT32BE = 5
+
+PA_STREAM_DONT_MOVE = 0x0200
+PA_STREAM_PEAK_DETECT = 0x0800
+PA_STREAM_ADJUST_LATENCY = 0x2000
+PA_STREAM_DONT_INHIBIT_AUTO_SUSPEND = 0x8000
+
+def c_enum_map(**values):
+	return dict((v, force_str(k)) for k,v in values.items())
+
+_globals = globals().copy()
+_pa_ev_type = dict(
+	(force_str(k), _globals['PA_SUBSCRIPTION_EVENT_{}'.format(k.upper())])
+	for k in 'new change remove'.split() )
+_pa_ev_fac, _pa_ev_mask = dict(), dict()
+for k, n in _globals.items():
+	if k.startswith('PA_SUBSCRIPTION_EVENT_'):
+		if k.endswith('_MASK'): continue
+		k = force_str(k[22:].lower())
+		if k in _pa_ev_type: continue
+		assert n & PA_SUBSCRIPTION_EVENT_FACILITY_MASK == n, [k, n]
+		_pa_ev_fac[k] = n
+	elif k.startswith('PA_SUBSCRIPTION_MASK_'):
+		_pa_ev_mask[force_str(k[21:].lower())] = n
+
+PA_EVENT_TYPE_MAP = c_enum_map(**_pa_ev_type)
+PA_EVENT_FACILITY_MAP = c_enum_map(**_pa_ev_fac)
+PA_EVENT_MASK_MAP = c_enum_map(**_pa_ev_mask)
+del _globals, _pa_ev_type, _pa_ev_fac, _pa_ev_mask
+
+PA_UPDATE_MAP = c_enum_map(set=0, merge=1, replace=2)
+PA_PORT_AVAILABLE_MAP = c_enum_map(unknown=0, no=1, yes=2)
+PA_DIRECTION_MAP = c_enum_map(unknown=0, output=1, input=2)
+
 # These are defined separately as
 #  pa_sink_state / pa_source_state, but seem to match.
-PA_OBJ_STATE_MAP = dict( (v,k)
-	for k,v in dict(invalid=-1, running=0, idle=1, suspended=2).items() )
+PA_OBJ_STATE_MAP = c_enum_map(invalid=-1, running=0, idle=1, suspended=2)
 
 
 class PA_MAINLOOP(Structure): pass
@@ -111,29 +151,26 @@ class PA_SAMPLE_SPEC(Structure):
 		('channels', c_uint32)
 	]
 
-
 class PA_CHANNEL_MAP(Structure):
 	_fields_ = [
 		('channels', c_uint8),
 		('map', c_int * PA_CHANNELS_MAX)
 	]
 
-
 class PA_CVOLUME(Structure):
 	_fields_ = [
 		('channels', c_uint8),
 		('values', c_uint32 * PA_CHANNELS_MAX)
 	]
 
-
 class PA_PORT_INFO(Structure):
 	_fields_ = [
 		('name', c_char_p),
 		('description', c_char_p),
 		('priority', c_uint32),
+		('available', c_int),
 	]
 
-
 class PA_SINK_INPUT_INFO(Structure):
 	_fields_ = [
 		('index', c_uint32),
@@ -155,7 +192,6 @@ class PA_SINK_INPUT_INFO(Structure):
 		('volume_writable', c_int),
 	]
 
-
 class PA_SINK_INFO(Structure):
 	_fields_ = [
 		('name', c_char_p),
@@ -182,7 +218,6 @@ class PA_SINK_INFO(Structure):
 		('active_port', POINTER(PA_PORT_INFO)),
 	]
 
-
 class PA_SOURCE_OUTPUT_INFO(Structure):
 	_fields_ = [
 		('index', c_uint32),
@@ -204,7 +239,6 @@ class PA_SOURCE_OUTPUT_INFO(Structure):
 		('volume_writable', c_int),
 	]
 
-
 class PA_SOURCE_INFO(Structure):
 	_fields_ = [
 		('name', c_char_p),
@@ -231,16 +265,15 @@ class PA_SOURCE_INFO(Structure):
 		('active_port', POINTER(PA_PORT_INFO)),
 	]
 
-
 class PA_CLIENT_INFO(Structure):
 	_fields_ = [
 		('index', c_uint32),
 		('name', c_char_p),
 		('owner_module', c_uint32),
 		('driver', c_char_p),
+		('proplist', POINTER(PA_PROPLIST)),
 	]
 
-
 class PA_SERVER_INFO(Structure):
 	_fields_ = [
 		('user_name', c_char_p),
@@ -254,7 +287,6 @@ class PA_SERVER_INFO(Structure):
 		('channel_map', PA_CHANNEL_MAP),
 	]
 
-
 class PA_CARD_PROFILE_INFO(Structure):
 	_fields_ = [
 		('name', c_char_p),
@@ -262,8 +294,23 @@ class PA_CARD_PROFILE_INFO(Structure):
 		('n_sinks', c_uint32),
 		('n_sources', c_uint32),
 		('priority', c_uint32),
+		('available', c_int),
 	]
 
+# Extends PA_PORT_INFO with a few card-specific things
+class PA_CARD_PORT_INFO(Structure):
+	_fields_ = [
+		('name', c_char_p),
+		('description', c_char_p),
+		('priority', c_uint32),
+		('available', c_int),
+		('direction', c_int),
+		('n_profiles', c_uint32),
+		('profiles', c_void_p), # use profiles2
+		('proplist', POINTER(PA_PROPLIST)),
+		('latency_offset', c_int64),
+		('profiles2', POINTER(POINTER(PA_CARD_PROFILE_INFO))),
+	]
 
 class PA_CARD_INFO(Structure):
 	_fields_ = [
@@ -272,12 +319,15 @@ class PA_CARD_INFO(Structure):
 		('owner_module', c_uint32),
 		('driver', c_char_p),
 		('n_profiles', c_uint32),
-		('profiles', POINTER(PA_CARD_PROFILE_INFO)),
-		('active_profile', POINTER(PA_CARD_PROFILE_INFO)),
+		('profiles', c_void_p), # use profiles2 / active_profile2
+		('active_profile', c_void_p),
 		('proplist', POINTER(PA_PROPLIST)),
+		('n_ports', c_uint32),
+		('ports', POINTER(POINTER(PA_CARD_PORT_INFO))),
+		('profiles2', POINTER(POINTER(PA_CARD_PROFILE_INFO))),
+		('active_profile2', POINTER(PA_CARD_PROFILE_INFO)),
 	]
 
-
 class PA_MODULE_INFO(Structure):
 	_fields_ = [
 		('index', c_uint32),
@@ -288,6 +338,24 @@ class PA_MODULE_INFO(Structure):
 		('proplist', POINTER(PA_PROPLIST)),
 	]
 
+class PA_EXT_STREAM_RESTORE_INFO(Structure):
+	_fields_ = [
+		('name', c_char_p),
+		('channel_map', PA_CHANNEL_MAP),
+		('volume', PA_CVOLUME),
+		('device', c_char_p),
+		('mute', c_int),
+	]
+
+class PA_BUFFER_ATTR(Structure):
+	_fields_ = [
+		('maxlength', c_uint32),
+		('tlength', c_uint32),
+		('prebuf', c_uint32),
+		('minreq', c_uint32),
+		('fragsize', c_uint32),
+	]
+
 
 class POLLFD(Structure):
 	_fields_ = [
@@ -309,7 +377,7 @@ PA_SIGNAL_CB_T = CFUNCTYPE(c_void_p,
 	c_int,
 	c_void_p)
 
-PA_STATE_CB_T = CFUNCTYPE(c_int,
+PA_STATE_CB_T = CFUNCTYPE(c_void_p,
 	POINTER(PA_CONTEXT),
 	c_void_p)
 
@@ -324,25 +392,25 @@ PA_SERVER_INFO_CB_T = CFUNCTYPE(c_void_p,
 	POINTER(PA_SERVER_INFO),
 	c_void_p)
 
-PA_SINK_INPUT_INFO_CB_T = CFUNCTYPE(c_int,
+PA_SINK_INPUT_INFO_CB_T = CFUNCTYPE(c_void_p,
 	POINTER(PA_CONTEXT),
 	POINTER(PA_SINK_INPUT_INFO),
 	c_int,
 	c_void_p)
 
-PA_SINK_INFO_CB_T = CFUNCTYPE(c_int,
+PA_SINK_INFO_CB_T = CFUNCTYPE(c_void_p,
 	POINTER(PA_CONTEXT),
 	POINTER(PA_SINK_INFO),
 	c_int,
 	c_void_p)
 
-PA_SOURCE_OUTPUT_INFO_CB_T = CFUNCTYPE(c_int,
+PA_SOURCE_OUTPUT_INFO_CB_T = CFUNCTYPE(c_void_p,
 	POINTER(PA_CONTEXT),
 	POINTER(PA_SOURCE_OUTPUT_INFO),
 	c_int,
 	c_void_p)
 
-PA_SOURCE_INFO_CB_T = CFUNCTYPE(c_int,
+PA_SOURCE_INFO_CB_T = CFUNCTYPE(c_void_p,
 	POINTER(PA_CONTEXT),
 	POINTER(PA_SOURCE_INFO),
 	c_int,
@@ -362,13 +430,24 @@ PA_CONTEXT_SUCCESS_CB_T = CFUNCTYPE(c_void_p,
 	c_int,
 	c_void_p)
 
-PA_CARD_INFO_CB_T = CFUNCTYPE(None,
+PA_EXT_STREAM_RESTORE_TEST_CB_T = CFUNCTYPE(c_void_p,
+	POINTER(PA_CONTEXT),
+	c_uint32,
+	c_void_p)
+
+PA_EXT_STREAM_RESTORE_READ_CB_T = CFUNCTYPE(c_void_p,
+	POINTER(PA_CONTEXT),
+	POINTER(PA_EXT_STREAM_RESTORE_INFO),
+	c_int,
+	c_void_p)
+
+PA_CARD_INFO_CB_T = CFUNCTYPE(c_void_p,
 	POINTER(PA_CONTEXT),
 	POINTER(PA_CARD_INFO),
 	c_int,
 	c_void_p)
 
-PA_MODULE_INFO_CB_T = CFUNCTYPE(None,
+PA_MODULE_INFO_CB_T = CFUNCTYPE(c_void_p,
 	POINTER(PA_CONTEXT),
 	POINTER(PA_MODULE_INFO),
 	c_int,
@@ -380,6 +459,15 @@ PA_SUBSCRIBE_CB_T = CFUNCTYPE(c_void_p,
 	c_int,
 	c_void_p)
 
+PA_STREAM_REQUEST_CB_T = CFUNCTYPE(c_void_p,
+	POINTER(PA_STREAM),
+	c_int,
+	c_void_p)
+
+PA_STREAM_NOTIFY_CB_T = CFUNCTYPE(c_void_p,
+	POINTER(PA_STREAM),
+	c_void_p)
+
 
 class LibPulse(object):
 
@@ -411,14 +499,21 @@ class LibPulse(object):
 		pa_context_connect=([POINTER(PA_CONTEXT), c_str_p, c_int, POINTER(c_int)], 'int_check_ge0'),
 		pa_context_get_state=([POINTER(PA_CONTEXT)], c_int),
 		pa_context_disconnect=[POINTER(PA_CONTEXT)],
+		pa_context_unref=[POINTER(PA_CONTEXT)],
 		pa_context_drain=( 'pa_op',
 			[POINTER(PA_CONTEXT), PA_CONTEXT_DRAIN_CB_T, c_void_p] ),
+		pa_context_set_default_sink=( 'pa_op',
+			[POINTER(PA_CONTEXT), c_str_p, PA_CONTEXT_SUCCESS_CB_T, c_void_p] ),
+		pa_context_set_default_source=( 'pa_op',
+			[POINTER(PA_CONTEXT), c_str_p, PA_CONTEXT_SUCCESS_CB_T, c_void_p] ),
 		pa_context_get_sink_input_info_list=( 'pa_op',
 			[POINTER(PA_CONTEXT), PA_SINK_INPUT_INFO_CB_T, c_void_p] ),
 		pa_context_get_sink_input_info=( 'pa_op',
 			[POINTER(PA_CONTEXT), c_uint32, PA_SINK_INPUT_INFO_CB_T, c_void_p] ),
 		pa_context_get_sink_info_list=( 'pa_op',
 			[POINTER(PA_CONTEXT), PA_SINK_INFO_CB_T, c_void_p] ),
+		pa_context_get_sink_info_by_name=( 'pa_op',
+			[POINTER(PA_CONTEXT), c_str_p, PA_SINK_INFO_CB_T, c_void_p] ),
 		pa_context_get_sink_info_by_index=( 'pa_op',
 			[POINTER(PA_CONTEXT), c_uint32, PA_SINK_INFO_CB_T, c_void_p] ),
 		pa_context_set_sink_mute_by_index=( 'pa_op',
@@ -449,6 +544,8 @@ class LibPulse(object):
 			[POINTER(PA_CONTEXT), c_uint32, PA_CONTEXT_SUCCESS_CB_T, c_void_p] ),
 		pa_context_get_source_info_list=( 'pa_op',
 			[POINTER(PA_CONTEXT), PA_SOURCE_INFO_CB_T, c_void_p] ),
+		pa_context_get_source_info_by_name=( 'pa_op',
+			[POINTER(PA_CONTEXT), c_str_p, PA_SOURCE_INFO_CB_T, c_void_p] ),
 		pa_context_get_source_info_by_index=( 'pa_op',
 			[POINTER(PA_CONTEXT), c_uint32, PA_SOURCE_INFO_CB_T, c_void_p] ),
 		pa_context_set_source_volume_by_index=( 'pa_op',
@@ -465,9 +562,11 @@ class LibPulse(object):
 			[POINTER(PA_CONTEXT), c_uint32, PA_CLIENT_INFO_CB_T, c_void_p] ),
 		pa_context_get_server_info=( 'pa_op',
 			[POINTER(PA_CONTEXT), PA_SERVER_INFO_CB_T, c_void_p] ),
-		pa_operation_unref=([POINTER(PA_OPERATION)], c_int),
+		pa_operation_unref=[POINTER(PA_OPERATION)],
 		pa_context_get_card_info_by_index=( 'pa_op',
 			[POINTER(PA_CONTEXT), c_uint32, PA_CARD_INFO_CB_T, c_void_p] ),
+		pa_context_get_card_info_by_name=( 'pa_op',
+			[POINTER(PA_CONTEXT), c_str_p, PA_CARD_INFO_CB_T, c_void_p] ),
 		pa_context_get_card_info_list=( 'pa_op',
 			[POINTER(PA_CONTEXT), PA_CARD_INFO_CB_T, c_void_p] ),
 		pa_context_set_card_profile_by_index=( 'pa_op',
@@ -482,16 +581,51 @@ class LibPulse(object):
 			[POINTER(PA_CONTEXT), c_uint32, PA_CONTEXT_SUCCESS_CB_T, c_void_p] ),
 		pa_context_subscribe=( 'pa_op',
 			[POINTER(PA_CONTEXT), c_int, PA_CONTEXT_SUCCESS_CB_T, c_void_p] ),
+		pa_ext_stream_restore_test=( 'pa_op',
+			[POINTER(PA_CONTEXT), PA_EXT_STREAM_RESTORE_TEST_CB_T, c_void_p] ),
+		pa_ext_stream_restore_read=( 'pa_op',
+			[POINTER(PA_CONTEXT), PA_EXT_STREAM_RESTORE_READ_CB_T, c_void_p] ),
+		pa_ext_stream_restore_write=( 'pa_op', [
+			POINTER(PA_CONTEXT), c_int, POINTER(PA_EXT_STREAM_RESTORE_INFO),
+			c_uint, c_int, PA_CONTEXT_SUCCESS_CB_T, c_void_p ] ),
+		pa_ext_stream_restore_delete=( 'pa_op',
+			[POINTER(PA_CONTEXT), POINTER(c_char_p), PA_CONTEXT_SUCCESS_CB_T, c_void_p] ),
 		pa_context_set_subscribe_callback=[POINTER(PA_CONTEXT), PA_SUBSCRIBE_CB_T, c_void_p],
 		pa_proplist_iterate=([POINTER(PA_PROPLIST), POINTER(c_void_p)], c_str_p),
 		pa_proplist_gets=([POINTER(PA_PROPLIST), c_str_p], c_str_p),
-		pa_channel_map_snprint=([c_str_p, c_int, POINTER(PA_CHANNEL_MAP)], c_str_p) )
+		pa_channel_map_init_mono=(
+			[POINTER(PA_CHANNEL_MAP)], (POINTER(PA_CHANNEL_MAP), 'not_null') ),
+		pa_channel_map_init_stereo=(
+			[POINTER(PA_CHANNEL_MAP)], (POINTER(PA_CHANNEL_MAP), 'not_null') ),
+		pa_channel_map_snprint=([c_str_p, c_int, POINTER(PA_CHANNEL_MAP)], c_str_p),
+		pa_channel_map_parse=(
+			[POINTER(PA_CHANNEL_MAP), c_str_p], (POINTER(PA_CHANNEL_MAP), 'not_null') ),
+		pa_proplist_from_string=([c_str_p], POINTER(PA_PROPLIST)),
+		pa_proplist_free=[POINTER(PA_PROPLIST)],
+		pa_stream_new_with_proplist=(
+			[ POINTER(PA_CONTEXT), c_str_p,
+				POINTER(PA_SAMPLE_SPEC), POINTER(PA_CHANNEL_MAP), POINTER(PA_PROPLIST) ],
+			POINTER(PA_STREAM) ),
+		pa_stream_set_monitor_stream=([POINTER(PA_STREAM), c_uint32], 'int_check_ge0'),
+		pa_stream_set_read_callback=[POINTER(PA_STREAM), PA_STREAM_REQUEST_CB_T, c_void_p],
+		pa_stream_connect_record=(
+			[POINTER(PA_STREAM), c_str_p, POINTER(PA_BUFFER_ATTR), c_int], 'int_check_ge0' ),
+		pa_stream_unref=[POINTER(PA_STREAM)],
+		pa_stream_peek=(
+			[POINTER(PA_STREAM), POINTER(c_void_p), POINTER(c_int)], 'int_check_ge0' ),
+		pa_stream_drop=([POINTER(PA_STREAM)], 'int_check_ge0'),
+		pa_stream_disconnect=([POINTER(PA_STREAM)], 'int_check_ge0'),
+		pa_context_play_sample=( 'pa_op',
+			[POINTER(PA_CONTEXT), c_str_p, c_str_p, c_uint32, PA_CONTEXT_SUCCESS_CB_T, c_void_p] ),
+		pa_context_play_sample_with_proplist=( 'pa_op',
+			[ POINTER(PA_CONTEXT), c_str_p, c_str_p, c_uint32,
+				POINTER(PA_PROPLIST), PA_CONTEXT_SUCCESS_CB_T, c_void_p ] ) )
 
 	class CallError(Exception): pass
 
 
 	def __init__(self):
-		p = CDLL('libpulse.so.0')
+		p = CDLL(ctypes.util.find_library('libpulse') or 'libpulse.so.0')
 
 		self.funcs = dict()
 		for k, spec in self.func_defs.items():
@@ -515,7 +649,8 @@ class LibPulse(object):
 		elif not func.restype: func.restype, res_proc = res_proc, None
 
 		def _wrapper(*args):
-			# print('libpulse call:', func_name, args, file=sys.stderr, flush=True)
+			# print('libpulse call:', func_name, args, file=sys.stderr)
+			# sys.stderr.flush()
 			res = func(*args)
 			if isinstance(res_proc, str):
 				assert res_proc in ['int_check_ge0', 'pa_op', 'not_null']
@@ -523,9 +658,9 @@ class LibPulse(object):
 						or (res_proc == 'pa_op' and not res)\
 						or (res_proc == 'not_null' and not res):
 					err = [func_name, args, res]
-					if args and isinstance(args[0].contents, PA_CONTEXT):
+					if args and isinstance(getattr(args[0], 'contents', None), PA_CONTEXT):
 						errno_ = self.context_errno(args[0])
-						err.append(self.strerror(errno_))
+						err.append('{} [pulse errno {}]'.format(self.strerror(errno_), errno_))
 					else: err.append('Return value check failed: {}'.format(res_proc))
 					raise self.CallError(*err)
 			elif res_proc: res = res_proc(res)
